# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["Plugin"]

import typing as t
from collections import defaultdict

import hikari

from lightbulb.utils import data_store

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import checks as checks_
    from lightbulb import commands
    from lightbulb import events

ListenerT = t.TypeVar("ListenerT", bound=t.Callable[..., t.Coroutine[t.Any, t.Any, None]])
ErrorHandlerT = t.TypeVar(
    "ErrorHandlerT",
    bound=t.Callable[..., t.Coroutine[t.Any, t.Any, t.Optional[bool]]],
)
RemoveHookT = t.TypeVar("RemoveHookT", bound=t.Callable[..., t.Union[t.Coroutine[t.Any, t.Any, None], None]])


class Plugin:
    """
    Container class for commands and listeners that can be loaded and unloaded from the bot
    to allow for hot-swapping of commands.

    Args:
        name (:obj:`str`): The name of the plugin.
        description (Optional[:obj:`str`]): Description of the plugin. Defaults to ``None``.
        include_datastore (:obj:`bool`): Whether or not to create a :obj:`~.utils.data_store.DataStore` instance
            internally for this plugin.
        default_enabled_guilds (UndefinedOr[Union[:obj:`int`, Sequence[:obj:`int`]]]): The guilds to create application
            commands registered to this plugin in by default. This overrides :obj:`~.app.BotApp.default_enabled_guilds`
            but is overridden by :obj:`~.commands.base.CommandLike.guilds`.

    .. versionadded:: 2.1.2
        ``default_enabled_guilds`` arg.
    """

    __slots__ = (
        "name",
        "description",
        "default_enabled_guilds",
        "_d",
        "_raw_commands",
        "_all_commands",
        "_listeners",
        "_app",
        "_checks",
        "_error_handler",
        "_remove_hook",
    )

    def __init__(
        self,
        name: str,
        description: t.Optional[str] = None,
        include_datastore: bool = False,
        default_enabled_guilds: hikari.UndefinedOr[t.Union[int, t.Sequence[int]]] = hikari.UNDEFINED,
    ) -> None:
        self.name: str = name
        """The plugin's name."""
        self.description: str = description or ""
        """The plugin's description."""
        self.default_enabled_guilds: hikari.UndefinedOr[t.Sequence[int]] = (
            (default_enabled_guilds,) if isinstance(default_enabled_guilds, int) else default_enabled_guilds
        )
        """The guilds that application commands registered to this plugin will be created in by default."""

        self._d: t.Optional[data_store.DataStore] = None
        if include_datastore:
            self._d = data_store.DataStore()

        self._raw_commands: t.List[commands.base.CommandLike] = []
        self._all_commands: t.List[commands.base.Command] = []

        self._listeners: t.MutableMapping[
            t.Type[hikari.Event], t.List[t.Callable[[hikari.Event], t.Coroutine[t.Any, t.Any, None]]]
        ] = defaultdict(list)

        self._checks: t.List[t.Union[checks_.Check, checks_._ExclusiveCheck]] = []
        self._error_handler: t.Optional[
            t.Callable[[events.CommandErrorEvent], t.Coroutine[t.Any, t.Any, t.Optional[bool]]]
        ] = None
        self._remove_hook: t.Optional[t.Callable[[], t.Union[t.Coroutine[t.Any, t.Any, None], None]]] = None

        self._app: t.Optional[app_.BotApp] = None

    @property
    def d(self) -> data_store.DataStore:
        """
        A :obj:`~.utils.data_store.DataStore` instance enabling storage of custom data without subclassing.
        This will raise a :obj:`RuntimeError` unless you explicitly specify you want the data
        storage instance included by passing the kwarg ``include_datastore=True`` to the constructor.
        """
        if self._d is None:
            raise RuntimeError(
                "'Plugin.d' cannot be accessed unless 'include_datastore=True' was provided during instantiation"
            )

        return self._d

    @property
    def app(self) -> app_.BotApp:
        """The :obj:`~.app.BotApp` instance that the plugin is registered to."""
        if self._app is None:
            raise RuntimeError(
                "'Plugin.app' cannot be accessed before the plugin has been added to a 'BotApp' instance"
            )

        return self._app

    @app.setter
    def app(self, val: app_.BotApp) -> None:
        self._app = val
        # Commands need the BotApp instance in order to be instantiated
        # so we wait until the instance is injected in order to create the Command instanced
        self.create_commands()

    @property
    def bot(self) -> app_.BotApp:
        """Alias for :obj:`~Plugin.app`"""
        return self.app

    @property
    def raw_commands(self) -> t.List[commands.base.CommandLike]:
        """List of all the CommandLike objects registered to the plugin."""
        return self._raw_commands

    @property
    def all_commands(self) -> t.List[commands.base.Command]:
        """List of all created command objects registered to the plugin."""
        return self._all_commands

    @property
    def listeners(
        self,
    ) -> t.Mapping[t.Type[hikari.Event], t.Collection[t.Callable[[hikari.Event], t.Coroutine[t.Any, t.Any, None]]]]:
        """Mapping of event type to listeners registered to the plugin"""
        return self._listeners

    def create_commands(self) -> None:
        """
        Creates the command objects implemented by the :obj:`~.commands.base.CommandLike` objects registered
        to the plugin.

        Returns:
            ``None``
        """
        assert self._app is not None
        for command_like in self._raw_commands:
            commands_to_impl: t.Sequence[t.Type[commands.base.Command]] = getattr(
                command_like.callback, "__cmd_types__", []
            )
            for cmd_type in commands_to_impl:
                cmd = cmd_type(self._app, command_like)

                if cmd.is_subcommand:
                    continue

                cmd._validate_attributes()
                cmd.plugin = self
                self._all_commands.append(cmd)

    @t.overload
    def command(self, cmd_like: commands.base.CommandLike) -> commands.base.CommandLike:
        ...

    @t.overload
    def command(self) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
        ...

    def command(
        self, cmd_like: t.Optional[commands.base.CommandLike] = None
    ) -> t.Union[commands.base.CommandLike, t.Callable[[commands.base.CommandLike], commands.base.CommandLike]]:
        """
        Adds a :obj:`~.commands.base.CommandLike` object as a command to the plugin. This method can be used as a
        first or second order decorator, or called manually with the :obj:`~.commands.CommandLike` instance to
        add as a command.
        """
        if cmd_like is not None:
            self._raw_commands.append(cmd_like)
            if cmd_like.guilds is hikari.UNDEFINED:
                cmd_like.guilds = self.default_enabled_guilds
            return cmd_like

        def decorate(cmd_like_: commands.base.CommandLike) -> commands.base.CommandLike:
            self.command(cmd_like_)
            return cmd_like_

        return decorate

    @t.overload
    def listener(self, event: t.Type[hikari.Event], listener_func: ListenerT, *, bind: bool = False) -> ListenerT:
        ...

    @t.overload
    def listener(self, event: t.Type[hikari.Event], *, bind: bool = False) -> t.Callable[[ListenerT], ListenerT]:
        ...

    def listener(
        self,
        event: t.Type[hikari.Event],
        listener_func: t.Optional[ListenerT] = None,
        *,
        bind: bool = False,
    ) -> t.Union[ListenerT, t.Callable[[ListenerT], ListenerT]]:
        """
        Adds a listener function to the plugin. This method can be used as a second order decorator, or called
        manually with the event type and function to add to the plugin as a listener.

        Args:
            event (Type[:obj:`~hikari.events.base_events.Event`): Event that the listener is for.

        Keyword Args:
            bind (:obj:`bool`): Whether or not to bind the listener function to the plugin. If ``True``, the
                function will be converted into a bound method and so will be called with the plugin as the
                first argument, and the error event as the second argument. Defaults to ``False``.
        """
        if listener_func is not None:
            if bind:
                listener_func = listener_func.__get__(self)
            assert listener_func is not None
            self._listeners[event].append(listener_func)
            return listener_func

        def decorate(func: ListenerT) -> ListenerT:
            # TODO - allow getting event type from type hint
            if bind:
                func = func.__get__(self)
            self.listener(event, func)
            return func

        return decorate

    @t.overload
    def set_error_handler(self, func: ErrorHandlerT, *, bind: bool = False) -> ErrorHandlerT:
        ...

    @t.overload
    def set_error_handler(self, *, bind: bool = False) -> t.Callable[[ErrorHandlerT], ErrorHandlerT]:
        ...

    def set_error_handler(
        self,
        func: t.Optional[ErrorHandlerT] = None,
        *,
        bind: bool = False,
    ) -> t.Union[ErrorHandlerT, t.Callable[[ErrorHandlerT], ErrorHandlerT]]:
        """
        Sets the error handler function for the plugin. This method can be used as a second order decorator,
        or called manually with the event type and function to set the plugin's error handler to.

        Keyword Args:
            bind (:obj:`bool`): Whether or not to bind the error handler function to the plugin. If ``True``, the
                function will be converted into a bound method and so will be called with the plugin as the
                first argument, and the error event as the second argument. Defaults to ``False``.
        """
        if func is not None:
            if bind:
                func = func.__get__(self)
            assert func is not None
            self._error_handler = func
            return func

        def decorate(func_: ErrorHandlerT) -> ErrorHandlerT:
            if bind:
                func_ = func_.__get__(self)
            self._error_handler = func_
            return func_

        return decorate

    @t.overload
    def remove_hook(self, func: RemoveHookT, *, bind: bool = False) -> RemoveHookT:
        ...

    @t.overload
    def remove_hook(self, *, bind: bool = False) -> t.Callable[[RemoveHookT], RemoveHookT]:
        ...

    def remove_hook(
        self,
        func: t.Optional[RemoveHookT] = None,
        *,
        bind: bool = False,
    ) -> t.Union[RemoveHookT, t.Callable[[RemoveHookT], RemoveHookT]]:
        """
        Sets the remove hook function for the plugin. This method can be used as a second order decorator,
        or called manually with the function to set the plugin's remove hook to. The registered function will
        be called when the plugin is removed from the bot so may be useful for teardown.

        This function will be called **after** all the members of the plugin (listeners and commands) have already
        been removed from the bot.

        Keyword Args:
            bind (:obj:`bool`): Whether or not to bind the remove hook function to the plugin. If ``True``, the
                function will be converted into a bound method and so will be called with the plugin as an
                argument. Defaults to ``False``.
        """
        if func is not None:
            if bind:
                func = func.__get__(self)
            assert func is not None
            self._remove_hook = func
            return func

        def decorate(func_: RemoveHookT) -> RemoveHookT:
            if bind:
                func_ = func_.__get__(self)
            self._remove_hook = func_
            return func_

        return decorate

    def add_checks(self, *checks: t.Union[checks_.Check, checks_._ExclusiveCheck]) -> None:
        """
        Adds one or more checks to the plugin object. These checks will be run for
        all commands in the plugin.

        Args:
            *checks (:obj:`~.checks.Check`): Check object(s) to add to the command.

        Returns:
            ``None``
        """
        self._checks.extend(checks)
        for check in checks:
            check.add_to_object_hook(self)



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["BaseParser", "Parser"]

import abc
import inspect
import logging
import typing as t

import hikari

from lightbulb import commands
from lightbulb import context as context_
from lightbulb import errors
from lightbulb.commands.base import OptionLike
from lightbulb.commands.base import OptionModifier
from lightbulb.converters import CONVERTER_TYPE_MAPPING
from lightbulb.converters import BaseConverter

T = t.TypeVar("T")
_quotes = {
    '"': '"',
    "'": "'",
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}
_LOGGER = logging.getLogger("lightbulb.utils.parser")


class BaseParser(abc.ABC):
    __slots__ = ("ctx", "_buffer", "len", "options")

    len: int
    ctx: context_.base.Context
    _buffer: str
    options: t.List[OptionLike]

    def __init__(
        self,
        context: context_.base.Context,
        buffer: t.Optional[str] = None,
        options: t.Optional[t.List[OptionLike]] = None,
    ) -> None:
        self.ctx = context

        if buffer is None:
            if not isinstance(self.ctx, context_.prefix.PrefixContext):
                raise RuntimeError("Please provide the buffer to parse")  # todo: proper error

            message = self.ctx.event.message
            assert message.content is not None
            buffer = message.content[len(self.ctx.prefix) + len(self.ctx.invoked_with) :]

        self.buffer = buffer

        if options is not None:
            self.options = options
        else:
            self.options = list(self.ctx.command.options.values()) if self.ctx.command else []

    @property
    def buffer(self) -> str:
        return self._buffer

    @buffer.setter
    def buffer(self, val: str) -> None:
        self._buffer = val
        self.len = len(val)

    @abc.abstractmethod
    async def parse(self) -> t.Dict[str, t.Any]:
        ...


class Parser(BaseParser):
    __slots__ = ("_idx", "prev")

    def __init__(
        self,
        context: context_.base.Context,
        buffer: t.Optional[str] = None,
        options: t.Optional[t.List[OptionLike]] = None,
    ) -> None:
        self._idx = 0
        self.prev = 0
        super().__init__(context, buffer, options)

    @property
    def is_eof(self) -> bool:
        return self.idx >= self.len

    @property
    def idx(self) -> int:
        return self._idx

    @idx.setter
    def idx(self, val: int) -> None:
        self.prev = self._idx
        self._idx = val

    def undo(self) -> None:
        self.idx = self.prev
        return None

    def skip_ws(self) -> None:
        prev = self.idx
        if (char := self.get_current()) is not None and not char.isspace():
            return None
        while (char := self.get_char()) is not None and char.isspace():
            pass
        self.prev = prev

    def get_char(self) -> t.Optional[str]:
        self.idx += 1
        return self.get_current()

    def get_current(self) -> t.Optional[str]:
        return None if self.is_eof else self.buffer[self.idx]

    def get_previous(self) -> t.Optional[str]:
        return None if self.idx == 0 else self.buffer[self.idx - 1]

    def get_word(self) -> str:
        """Gets the next word, will return an empty string if EOF."""
        self.skip_ws()
        prev = self.idx
        while (char := self.get_char()) is not None and not char.isspace():
            pass
        self.prev = prev
        return self.buffer[prev : self.idx]

    def get_quoted_word(self) -> str:
        self.skip_ws()
        prev = self.idx
        if (closing := _quotes.get(t.cast(str, self.get_current()))) is None:
            return self.get_word()

        while (char := self.get_char()) is not None:
            if char == closing and self.get_previous() != "\\":
                break
        else:
            # EOF
            raise RuntimeError("expected a closing quote")  # TODO: raise proper error

        if (current := self.get_char()) is not None and not current.isspace():
            raise RuntimeError("expected a space after the closing quote")  # TODO: raise proper error

        self.prev = prev
        return self.buffer[prev + 1 : self.idx - 1].replace(f"\\{closing}", closing)

    def read_rest(self) -> str:
        self.skip_ws()
        self.idx = self.len
        return self.buffer[self.prev :]

    def get_option(self) -> t.Optional[commands.base.OptionLike]:
        if self.options:
            return self.options.pop(0)

        return None

    async def parse(self) -> t.Dict[str, t.Any]:
        ret = {}
        attachments = list(self.ctx.attachments) if isinstance(self.ctx, context_.prefix.PrefixContext) else []
        while option := self.get_option():
            if option.arg_type in (hikari.OptionType.ATTACHMENT, hikari.Attachment):
                if not attachments:
                    if not option.required:
                        ret[option.name] = option.default
                        continue
                    raise errors.MissingRequiredAttachmentArgument(
                        "Command invocation expects an attachment but none were found.", missing=option
                    )
                ret[option.name] = attachments.pop(0)
                continue

            _LOGGER.debug("Getting arg for %s with type %s", option.name, option.arg_type)

            if not (
                raw_arg := self.read_rest()
                if option.modifier is OptionModifier.CONSUME_REST
                else self.get_quoted_word()
            ):
                _LOGGER.debug("Arguments have exhausted")
                if option.required:
                    raise errors.NotEnoughArguments(
                        "Command invocation is missing one or more required arguments.",
                        missing=[option, *(o for o in self.options if o.required)],
                    )

                ret[option.name] = option.default
                continue

            _LOGGER.debug("Got raw arg %s", raw_arg)
            convert = self._greedy_convert if option.modifier is OptionModifier.GREEDY else self._try_convert
            await convert(raw_arg, option, ret)

        return ret

    async def _try_convert(self, raw: str, option: commands.base.OptionLike, out: t.Dict[str, t.Any]) -> None:
        try:
            arg = await self._convert(raw, option.arg_type)
        except Exception as e:
            _LOGGER.debug("Failed to convert", exc_info=e)
            if option.required:
                raise errors.ConverterFailure(
                    f"Conversion failed for option {option.name!r}", opt=option, raw=raw
                ) from e

            out[option.name] = option.default
            _LOGGER.debug("Option has a default value, shifting to the next parameter")
            self.undo()
        else:
            _LOGGER.debug("Successfully converted %s to %s", raw, arg)
            out[option.name] = arg

    async def _greedy_convert(self, raw: str, option: commands.base.OptionLike, out: t.Dict[str, t.Any]) -> None:
        out[option.name] = args = []
        _LOGGER.debug("Attempting to greedy convert %s to %s", raw, option.arg_type)
        while raw:
            try:
                arg = await self._convert(raw, option.arg_type)
            except Exception as e:
                _LOGGER.debug("Done greedy converting", exc_info=e)
                self.undo()
                break
            else:
                _LOGGER.debug("Appending %s", arg)
                args.append(arg)
                raw = self.get_quoted_word()

    async def _convert(
        self,
        value: str,
        callback_or_type: t.Union[
            t.Callable[[str], t.Union[T, t.Coroutine[t.Any, t.Any, T]]], t.Type[BaseConverter[T]]
        ],
    ) -> T:
        callback_or_type = CONVERTER_TYPE_MAPPING.get(callback_or_type, callback_or_type)
        _LOGGER.debug("Attempting to convert %s to %s", value, callback_or_type)
        conversion_func = callback_or_type
        if inspect.isclass(callback_or_type) and issubclass(callback_or_type, BaseConverter):
            conversion_func = callback_or_type(self.ctx).convert

        converted = conversion_func(value)  # type: ignore
        if inspect.iscoroutine(converted):
            assert isinstance(converted, t.Awaitable)
            converted = await converted

        converted = t.cast(T, converted)
        return converted



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["PrefixCommand", "PrefixSubCommand", "PrefixSubGroup", "PrefixCommandGroup", "PrefixGroupMixin"]

import abc
import typing as t

from multidict import CIMultiDict

from lightbulb import context as context_
from lightbulb import errors
from lightbulb.commands import base

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import plugins


class PrefixGroupMixin(abc.ABC):
    __slots__ = ()
    _plugin: t.Optional[plugins.Plugin]
    _subcommands: t.Dict[str, t.Union[PrefixSubGroup, PrefixSubCommand]]

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    def maybe_resolve_subcommand(
        self, arg_string: str
    ) -> t.Tuple[t.Optional[t.Union[PrefixSubGroup, PrefixSubCommand]], str]:
        if not arg_string:
            return None, ""

        maybe_subcmd, *remainder = arg_string.split(maxsplit=1)
        remainder = "".join(remainder)

        if (cmd := self._subcommands.get(maybe_subcmd)) is not None:
            return cmd, remainder
        return None, ""

    def create_subcommands(self, raw_cmds: t.Sequence[base.CommandLike], app: app_.BotApp) -> None:
        for raw_cmd in raw_cmds:
            impls: t.List[t.Type[base.Command]] = getattr(raw_cmd.callback, "__cmd_types__", [])
            for impl in impls:
                if issubclass(impl, (PrefixSubCommand, PrefixSubGroup)):
                    cmd = impl(app, raw_cmd)
                    cmd.parent = self  # type: ignore
                    cmd.plugin = self.plugin  # type: ignore

                    for name in [cmd.name, *cmd.aliases]:
                        if name in self._subcommands:
                            raise errors.CommandAlreadyExists(
                                f"A prefix subcommand with name or alias {name!r} already exists for group {self.name!r}"
                            )
                        self._subcommands[name] = cmd

    def recreate_subcommands(self, raw_cmds: t.Sequence[base.CommandLike], app: app_.BotApp) -> None:
        self._subcommands.clear()
        self.create_subcommands(raw_cmds, app)

    def get_subcommand(self, name: str) -> t.Optional[t.Union[PrefixSubGroup, PrefixSubCommand]]:
        return self._subcommands.get(name)

    @property
    def subcommands(self) -> t.Dict[str, t.Union[PrefixSubGroup, PrefixSubCommand]]:
        """Mapping of command name to command object containing the group's subcommands."""
        return self._subcommands

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        self._plugin = pl
        for command in self._subcommands.values():
            if isinstance(command, PrefixGroupMixin):
                command._set_plugin(pl)
            else:
                command.plugin = pl


class PrefixCommand(base.Command):
    """
    An implementation of :obj:`~.commands.base.Command` representing a prefix command.
    """

    __slots__ = ()

    @property
    def signature(self) -> str:
        sig = self.qualname
        if self.options:
            sig += f" {' '.join(f'<{o.name}>' if o.required else f'[{o.name}={o.default}]' for o in self.options.values())}"
        return sig

    async def invoke(self, context: context_.base.Context, **kwargs: t.Any) -> None:
        context._invoked = self

        await self._evaluate_max_concurrency(context)
        try:
            await self.evaluate_checks(context)
            await self.evaluate_cooldowns(context)
            assert isinstance(context, context_.prefix.PrefixContext)
            context._options = await context._parser.parse()
            await self(context, **kwargs)
        except Exception:
            raise
        finally:
            self._release_max_concurrency(context)

    def _validate_attributes(self) -> None:
        if " " in self.name:
            raise ValueError(f"Prefix command {self.name!r}: name cannot contain spaces") from None


class PrefixSubCommand(PrefixCommand, base.SubCommandTrait):
    """
    Class representing a prefix subcommand.
    """

    __slots__ = ()

    @property
    def qualname(self) -> str:
        assert self.parent is not None
        return f"{self.parent.qualname} {self.name}"

    async def invoke(self, context: context_.base.Context, *, _arg_buffer: str = "", **kwargs: t.Any) -> None:
        context._invoked = self
        assert isinstance(context, context_.prefix.PrefixContext)
        context._parser = type(context._parser)(context, _arg_buffer)
        context._parser.options = list(self.options.values())
        await super().invoke(context, **kwargs)


class PrefixSubGroup(PrefixCommand, PrefixGroupMixin, base.SubCommandTrait):
    """
    Class representing a prefix subgroup of commands.
    """

    __slots__ = ("_raw_subcommands", "_subcommands")

    def __init__(self, app: app_.BotApp, initialiser: base.CommandLike) -> None:
        super().__init__(app, initialiser)
        self._raw_subcommands = initialiser.subcommands
        initialiser.subcommands = (
            initialiser.subcommands.add_parent(self)  # type: ignore
            if isinstance(initialiser.subcommands, base._SubcommandListProxy)  # type: ignore
            else base._SubcommandListProxy(initialiser.subcommands, parent=self)
        )
        self._subcommands = {} if not app._case_insensitive_prefix_commands else CIMultiDict()  # type: ignore
        self.create_subcommands(self._raw_subcommands, app)

    @property
    def qualname(self) -> str:
        assert self.parent is not None
        return f"{self.parent.qualname} {self.name}"

    async def invoke(self, context: context_.base.Context, *, _arg_buffer: str = "", **kwargs: t.Any) -> None:
        context._invoked = self
        subcmd, remainder = self.maybe_resolve_subcommand(_arg_buffer)
        if subcmd is not None:
            await subcmd.invoke(context, _arg_buffer=remainder, **kwargs)
            return

        assert isinstance(context, context_.prefix.PrefixContext)
        context._parser = type(context._parser)(context, _arg_buffer)
        context._parser.options = list(self.options.values())
        await super().invoke(context, **kwargs)

    def _validate_attributes(self) -> None:
        super()._validate_attributes()
        for command in self._subcommands.values():
            command._validate_attributes()

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        PrefixGroupMixin._set_plugin(self, pl)


class PrefixCommandGroup(PrefixCommand, PrefixGroupMixin):
    """
    Class representing a prefix command group.
    """

    __slots__ = ("_raw_subcommands", "_subcommands")

    def __init__(self, app: app_.BotApp, initialiser: base.CommandLike) -> None:
        super().__init__(app, initialiser)
        self._raw_subcommands = initialiser.subcommands
        initialiser.subcommands = (
            initialiser.subcommands.add_parent(self)  # type: ignore
            if isinstance(initialiser.subcommands, base._SubcommandListProxy)  # type: ignore
            else base._SubcommandListProxy(initialiser.subcommands, parent=self)
        )
        self._subcommands = {} if not app._case_insensitive_prefix_commands else CIMultiDict()  # type: ignore
        self.create_subcommands(self._raw_subcommands, app)

    async def invoke(self, context: context_.base.Context, **kwargs: t.Any) -> None:
        context._invoked = self
        assert isinstance(context, context_.prefix.PrefixContext) and context.event.message.content is not None

        subcmd, remainder = self.maybe_resolve_subcommand(
            context.event.message.content[len(context.prefix) + len(context.invoked_with) :].strip()
        )
        if subcmd is not None:
            await subcmd.invoke(context, _arg_buffer=remainder)
            return

        await super().invoke(context, **kwargs)

    def _validate_attributes(self) -> None:
        super()._validate_attributes()
        for command in self._subcommands.values():
            command._validate_attributes()

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        PrefixGroupMixin._set_plugin(self, pl)



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
"""
An implementation of simple repeating asyncio tasks.

.. versionadded:: 2.2.0

Setup
-----

In order for tasks to function correctly you must first call :meth:`load` after initialising
your instance of the :obj:`~lightbulb.app.BotApp` class. See below:

.. code-block:: python

    import lightbulb
    from lightbulb.ext import tasks

    app = lightbulb.BotApp(...)
    tasks.load(app)

Creating Tasks
--------------

Tasks are created using the :obj:`~task` decorator.

.. code-block:: python

    from lightbulb.ext import tasks

    # s=30 means this will run every 30 seconds
    # you can also use (m)inutes, (h)ours, and (d)ays
    @tasks.task(s=30)
    async def print_every_30_seconds():
        print("Task called")

    # Instead of having to call .start() manually, you can pass auto_start=True
    # into the task decorator if you wish
    print_every_30_seconds.start()

See the :obj:`~task` decorator api reference for valid kwargs you can pass.

API Reference
-------------
"""
from __future__ import annotations

__all__ = ["load", "task", "wait_until_started", "Task", "Trigger", "UniformTrigger", "CronTrigger", "triggers"]

import asyncio
import functools
import inspect
import logging
import typing as t

import hikari

from . import triggers
from .triggers import *

if t.TYPE_CHECKING:
    import lightbulb

_LOGGER = logging.getLogger("lightbulb.ext.tasks")
TaskCallbackT = t.TypeVar("TaskCallbackT", bound=t.Callable[..., t.Union[t.Any, t.Coroutine[t.Any, t.Any, t.Any]]])
TaskErrorHandlerT = t.TypeVar(
    "TaskErrorHandlerT", bound=t.Callable[..., t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]]
)


class _BindableObjectWithCallback:
    __slots__ = ("_callback", "_bound")

    def __init__(self, callback: t.Callable[..., t.Any]) -> None:
        self._callback = callback
        self._bound: bool = False

    def __get__(self, instance: t.Any, _: t.Type[t.Any]) -> _BindableObjectWithCallback:
        if self._bound or instance is None:
            return self
        self._callback = functools.partial(self._callback, instance)
        return self

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self._callback(*args, **kwargs)


class Task(_BindableObjectWithCallback):
    """
    Class representing a repeating task.

    Args:
        callback: The function that will be executed every time the task is run.
        trigger (:obj:`.triggers.Trigger`): The task trigger to use.
        auto_start (:obj:`bool`): Whether the task will be automatically started when instantiated.
        max_consecutive_failures (:obj:`int`): The number of consecutive task failures that will be ignored
            before the task is cancelled.
        max_executions (Optional[:obj:`int`]): The maximum number of times that the task will be executed. If ``None``,
            the task will run indefinitely.
        pass_app (:obj:`bool`): Whether the :obj:`lightbulb.app.BotApp` object will be passed into the task upon
            execution.
    """

    __slots__ = (
        "_callback",
        "_trigger",
        "_stopped",
        "_task",
        "_error_handler",
        "_consecutive_failures",
        "_max_consecutive_failures",
        "_max_executions",
        "_n_executions",
        "_pass_app",
        "_wait_before_execution",
    )
    _app: t.Optional[lightbulb.BotApp] = None
    _app_starting: asyncio.Event = asyncio.Event()
    _app_started: asyncio.Event = asyncio.Event()
    _tasks: t.List[Task] = []

    def __init__(
        self,
        callback: TaskCallbackT,
        trigger: triggers.Trigger,
        auto_start: bool,
        max_consecutive_failures: int,
        max_executions: t.Optional[int],
        pass_app: bool,
        wait_before_execution: hikari.UndefinedOr[bool],
    ) -> None:
        super().__init__(callback)
        self._trigger: triggers.Trigger = trigger
        self._stopped: bool = False
        self._task: t.Optional[asyncio.Task[t.Any]] = None
        self._error_handler: t.Optional[
            t.Callable[[BaseException], t.Union[t.Any, t.Coroutine[t.Any, t.Any, t.Any]]]
        ] = None
        self._max_consecutive_failures: int = max_consecutive_failures
        self._consecutive_failures: int = 0
        self._max_executions: t.Optional[int] = max_executions
        self._n_executions: int = 0
        self._pass_app: bool = pass_app
        self._wait_before_execution: bool = (
            wait_before_execution if wait_before_execution is not hikari.UNDEFINED else trigger.wait_before_execution
        )

        if auto_start:
            self.start()

    def __get__(self, instance: t.Any, owner: t.Type[t.Any]) -> Task:
        # We override this to keep linters happy
        super().__get__(instance, owner)
        return self

    @property
    def __name__(self) -> str:
        return getattr(
            self._callback.func if isinstance(self._callback, functools.partial) else self._callback,
            "__name__",
            "__unknown_name__",
        )

    @property
    def _is_event_loop_running(self) -> bool:
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return False
        return True

    @classmethod
    async def _app_starting_listener(cls, _: hikari.StartingEvent) -> None:
        assert Task._app is not None
        Task._app_starting.set()
        for task_ in Task._tasks:
            if not task_.is_running:
                task_.start()

    @classmethod
    async def _app_started_listener(cls, _: hikari.StartedEvent) -> None:
        assert Task._app is not None
        Task._app_started.set()

    @classmethod
    async def _app_stopping_listener(cls, _: hikari.StoppingEvent) -> None:
        assert Task._app is not None
        Task._app_starting.clear()
        Task._app_started.clear()
        for task_ in Task._tasks:
            task_.stop()

    async def _loop(self) -> None:
        if self._wait_before_execution:
            await asyncio.sleep(self._trigger.get_interval())

        while not self._stopped:
            if self._max_executions is not None and self._n_executions >= self._max_executions:
                break

            _LOGGER.debug("Running task %r", self.__name__)
            self._n_executions += 1
            try:
                maybe_coro = self._callback(*([Task._app] if self._pass_app else []))
                if inspect.iscoroutine(maybe_coro):
                    await maybe_coro
                self._consecutive_failures = 0
            except Exception as e:
                out: t.Any = False

                if self._error_handler is not None:
                    out = self._error_handler(e)
                    if inspect.iscoroutine(out):
                        out = await out

                if not out:
                    self._consecutive_failures += 1
                    if self._consecutive_failures >= self._max_consecutive_failures:
                        _LOGGER.error(
                            "Task failed repeatedly and has been cancelled", exc_info=(type(e), e, e.__traceback__)
                        )
                        break

                    _LOGGER.error(
                        "Error occurred during task execution and was not handled. "
                        "Task will be cancelled after %s more consecutive failure(s)",
                        self._max_consecutive_failures - self._consecutive_failures,
                        exc_info=(type(e), e, e.__traceback__),
                    )

            await asyncio.sleep(self._trigger.get_interval())

        assert self._task is not None
        self._task.cancel()
        Task._tasks.remove(self)
        self._task = None
        self._stopped = True
        self._n_executions = 0
        _LOGGER.debug("Stopped task %r", self.__name__)

    @property
    def n_executions(self) -> int:
        """The number of times the task has been executed since being started."""
        return self._n_executions

    @property
    def is_running(self) -> bool:
        """Whether the task represented by this object is currently running or not."""
        return self._task is not None and not self._task.done()

    @t.overload
    def set_error_handler(self, func: TaskErrorHandlerT) -> TaskErrorHandlerT:
        ...

    @t.overload
    def set_error_handler(self) -> t.Callable[[TaskErrorHandlerT], TaskErrorHandlerT]:
        ...

    def set_error_handler(
        self, func: t.Optional[TaskErrorHandlerT] = None
    ) -> t.Union[TaskErrorHandlerT, t.Callable[[TaskErrorHandlerT], TaskErrorHandlerT]]:
        """
        Sets the function to use as the error handler for this task. This can be used as a first
        or second order decorator, or called with the function to set as the error handler.

        The error handler should return a boolean indicating whether the error could be handled.
        If ``False`` or a falsy value is returned, the related execution of the task will be considered
        to be a failure.
        """
        if func is not None:
            if isinstance(func, _BindableObjectWithCallback):
                return self._error_handler  # type: ignore[unreachable]
            self._error_handler = _BindableObjectWithCallback(func)
            return self._error_handler

        def decorate(func_: TaskErrorHandlerT) -> TaskErrorHandlerT:
            if isinstance(func_, _BindableObjectWithCallback):
                return func_  # type: ignore[unreachable]
            self._error_handler = _BindableObjectWithCallback(func_)
            return self._error_handler

        return decorate

    def start(self) -> None:
        """
        Start the task if the event loop has been established, or schedule the task to be started
        once the event loop has been established.

        Returns:
            ``None``
        """
        Task._tasks.append(self)
        if self._is_event_loop_running:
            _LOGGER.debug("Starting task %r", self.__name__)
            self._task = asyncio.create_task(self._loop())

    def stop(self) -> None:
        """
        Stop the task after the completion of the current iteration, if it was running.

        Returns:
            ``None``
        """
        if self.is_running:
            self._stopped = True

    def cancel(self) -> None:
        """
        Cancel the task if it was running.

        Returns:
            ``None``
        """
        if not self.is_running:
            return

        assert self._task is not None
        self._task.cancel()
        self._task = None
        self._stopped = True
        self._n_executions = 0


async def wait_until_started() -> None:
    """
    Wait until the bot has started. Note that :obj:`.load` **must** have been
    called in order for this to function.

    Roughly equivalent to:

    .. code-block:: python

        # Where app is an instance of 'lightbulb.BotApp'
        await app.wait_for(hikari.StartedEvent, timeout=None)

    Returns:
        ``None``
    """
    await Task._app_started.wait()


@t.overload
def task(
    *,
    s: float = 0,
    m: float = 0,
    h: float = 0,
    d: float = 0,
    auto_start: bool = False,
    max_consecutive_failures: int = 3,
    max_executions: t.Optional[int] = None,
    pass_app: bool = False,
    wait_before_execution: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
    cls: t.Type[Task] = Task,
) -> t.Callable[[TaskCallbackT], Task]:
    ...


@t.overload
def task(
    trigger: triggers.Trigger,
    /,
    *,
    auto_start: bool = False,
    max_consecutive_failures: int = 3,
    max_executions: t.Optional[int] = None,
    pass_app: bool = False,
    wait_before_execution: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
    cls: t.Type[Task] = Task,
) -> t.Callable[[TaskCallbackT], Task]:
    ...


@t.overload
def task(
    trigger: t.Type[triggers.UniformTrigger],
    /,
    *,
    s: float = 0,
    m: float = 0,
    h: float = 0,
    d: float = 0,
    auto_start: bool = False,
    max_consecutive_failures: int = 3,
    max_executions: t.Optional[int] = None,
    pass_app: bool = False,
    wait_before_execution: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
    cls: t.Type[Task] = Task,
) -> t.Callable[[TaskCallbackT], Task]:
    ...


def task(
    trigger: t.Optional[t.Union[triggers.Trigger, t.Type[triggers.Trigger]]] = None,
    /,
    *,
    s: float = 0,
    m: float = 0,
    h: float = 0,
    d: float = 0,
    auto_start: bool = False,
    max_consecutive_failures: int = 3,
    max_executions: t.Optional[int] = None,
    pass_app: bool = False,
    wait_before_execution: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
    cls: t.Type[Task] = Task,
) -> t.Callable[[TaskCallbackT], Task]:
    """
    Second order decorator to register a function as a repeating task. The decorated function can
    be a synchronous or asynchronous function, and any return value will be discarded.

    Args:
        trigger (Optional[:obj:`~.triggers.Trigger`]): Trigger to use to set the interval between task executions.
            If not provided, and interval values were provided then this defaults to :obj:`~.triggers.UniformTrigger`.

    Keyword Args:
        s (:obj:`float`): Number of seconds between task executions.
        m (:obj:`float`): Number of minutes between task executions.
        h (:obj:`float`): Number of hours between task executions.
        d (:obj:`float`): Number of days between task executions.
        auto_start (:obj:`bool`): Whether the task will be started automatically when created. If ``False``,
            :meth:`~Task.start` will have to be called manually in order to start the task's execution.
        max_consecutive_failures (:obj:`int`): The number of consecutive task failures that will be ignored
            before the task's execution is cancelled. Defaults to ``3``, minimum ``1``.
        max_executions (Optional[:obj:`int`]): The maximum number of times that the task will run. If ``None``, the
            task will run indefinitely.
        pass_app (:obj:`bool`): Whether the :obj:`lightbulb.app.BotApp` instance should be passed into the
            task on execution. Defaults to ``False``.
        wait_before_execution (UndefinedOr[:obj:`bool`]): Whether the task will wait its given interval before the first
            time the task is executed. Defaults to ``UNDEFINED`` (will use the trigger's default).
        cls (Type[:obj:`~Task`]): Task class to use.

    .. versionadded:: 2.2.2
        ``wait_before_execution`` kwarg.
    """

    def decorate(func: TaskCallbackT) -> Task:
        nonlocal trigger

        if any([s, m, h, d]):
            trigger = trigger or triggers.UniformTrigger
            assert inspect.isclass(trigger) and issubclass(trigger, triggers.UniformTrigger)
            return cls(
                func,
                (trigger or triggers.UniformTrigger)(s + (m * 60) + (h * 3600) + (d * 86400)),
                auto_start,
                max(max_consecutive_failures, 1),
                max_executions,
                pass_app,
                wait_before_execution,
            )

        if trigger is None:
            raise ValueError("Interval values were not provided and no trigger was passed")

        assert isinstance(trigger, triggers.Trigger)
        return cls(
            func, trigger, auto_start, max(max_consecutive_failures, 1), max_executions, pass_app, wait_before_execution
        )

    return decorate


def load(app: lightbulb.BotApp) -> None:
    """
    Add the task system to the bot, enabling tasks to be run
    once the bot has started.

    Args:
        app (:obj:`~lightbulb.app.BotApp`): :obj:`~lightbulb.app.BotApp` instance
            that tasks will be enabled for.

    Returns:
        ``None``
    """
    Task._app = app
    app.subscribe(hikari.StartingEvent, Task._app_starting_listener)
    app.subscribe(hikari.StartedEvent, Task._app_started_listener)
    app.subscribe(hikari.StoppingEvent, Task._app_stopping_listener)



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "ApplicationContext",
    "Context",
    "MessageContext",
    "OptionsProxy",
    "PrefixContext",
    "ResponseProxy",
    "SlashContext",
    "UserContext",
]

from lightbulb.context import base
from lightbulb.context import message
from lightbulb.context import prefix
from lightbulb.context import slash
from lightbulb.context import user
from lightbulb.context.base import *
from lightbulb.context.message import *
from lightbulb.context.prefix import *
from lightbulb.context.slash import *
from lightbulb.context.user import *
